LevelPolygon = Entity:extend()

require "entities/polygonSpecial"

--Class for level polygons

function LevelPolygon:new(x, y, inputPoints, attributes)
  Polygon.super.new(self, x, y)
  self.type = "levelPolygon"
  
  if attributes then
    self.attributes = attributes
  else
    self.attributes = {}
    self.attributes.isBackground  = false
    self.attributes.draw          = true
    self.attributes.textureName   = "stone"
    self.attributes.smooth        = false
    self.attributes.hasPhysics    = false
    self.attributes.hasGravity    = false
    self.attributes.springBack    = true
    self.attributes.rotateBack    = true
    self.attributes.protected     = false
  end
  
  createAttribute(self.attributes, "collidable", true)
  createAttribute(self.attributes, "priorityDraw", false)
  createAttribute(self.attributes, "slowMotion", false)
  
  
  self.attributes.ignore = false -- It's not recommended that you create an attribute manually, but I always want loaded polygons to not be ignored
  
  if inputPoints then
    self.points = inputPoints
    self.finalised = true
    self:makeStatic()
  else
    self.points = {}
    self.finalised = false
    self.points[1] = vector.create(x, y)
    self.numberOfPoints = 1
  end
  self.selected = false
  table.insert(drawTable, self)
end

function LevelPolygon:addPoint(x, y)
  local point = Point(x, y)
  local lastPoint = self.points[#self.points]
  
  
  --Limiter to prevent box2D error
  if point:distanceToTargetSqrd(lastPoint) > (0.005*64)^2 then
    table.insert(self.points, point)
    self.numberOfPoints = self.numberOfPoints + 1
    return point
  else
    return lastPoint
  end
end

function LevelPolygon:removePoint(index)
  if index then
    
    index = 2*index - 1
    
    table.remove(self.points, index)
    table.remove(self.points, index)
  
  else
    table.remove(self.points, self.numberOfPoints)
    self.numberOfPoints = self.numberOfPoints - 1
  end
end


function LevelPolygon:selectPoint(x, y, table)
  
  --returns the index of the closest vertex of a polygon to a given position
    
  if builder.selectedVertex then
    return builder.selectedVertex
  end
  
  
  local selectionDistance = 15/camera.zoom
  local selectionDistanceSrd = selectionDistance^2
  
  local closestDistanceSqrd = nil
  local closestPoint = nil
  
  if table == nil then
    table = self.points
  end
  
  for i, v in ipairs(table) do

    if i%2 == 0 then
      -- If even, continue
    else
      local pointX = table[i] + self.originalX
      local pointY = table[i+1] + self.originalY
      
      local pointDistanceSqrd = (x - pointX)^2 + (y - pointY)^2
      
      
      if closestDistanceSqrd == nil or pointDistanceSqrd < closestDistanceSqrd then
        closestDistanceSqrd = pointDistanceSqrd
        closestPoint = (i+1)/2
      end
    end
  end
  
  love.graphics.setColor(0,0,0)
  
  
  if closestDistanceSqrd <= selectionDistanceSrd then
    builder.selectedVertex = closestPoint
    return closestPoint
  else
    return nil
  end
end
  


function LevelPolygon:movePoint(x, y, i)
  -- i is the index of the vertex
  -- NOTE: polygons in this form are contained as a single table, odd values being the x and even values being the y.
  -- x = 2i - 1
  -- y = 2i
  self.points[2*i - 1] = x
  self.points[2*i] = y
end


function LevelPolygon:finalise()
  self.finalised = true
  local newPointFormat = {}
  for i, point in ipairs(self.points) do
    table.insert(newPointFormat, point.x)
    table.insert(newPointFormat, point.y)
  end
  self.points = {}
  self.points = newPointFormat
  
  
  self:makeStatic()
  --self.remove = true
  
  currentLevel.polygons[self] = self
  
end

function LevelPolygon:makeStatic()
  if self.finalised then    
    

    self.body = love.physics.newBody(world, 0, 0)
    self.shape = love.physics.newChainShape(true, self.points)
    self.fixture = love.physics.newFixture(self.body, self.shape)
    self.fixture:setFriction(0.8)
    loadingScreen()
    if self.attributes.draw then
      self:getCanvasSet()
    end
  else
    cout("Cannot make static, polygon is not finalised!")
  end
end



function LevelPolygon:remake()
  
  self:setAttributes()
  if self.attributes.draw then
    self:getCanvasSet()
  end
end


function LevelPolygon:setAttributes()

  if self.attributes.hasPhysics then
    local newCentre = self:getTopCorner()
    
    
    self.shape = nil
    self.body:destroy()
    self.body = love.physics.newBody(world, newCentre.x, newCentre.y, "dynamic")
    
    
    for i=1, #self.points, 2 do
      self.points[i] = self.points[i]
      self.points[i+1] = self.points[i+1]
    end
    
    local triangles = love.math.triangulate(self.points)
    self.triangleShapes = {}
    self.triangleFixtures = {}
    for i, triangle in ipairs(triangles) do
      self.triangleShapes[i] = love.physics.newPolygonShape(unpack(triangle))
      self.triangleFixtures[i] = love.physics.newFixture(self.body, self.triangleShapes[i])
      self.triangleFixtures[i]:setFriction(0.8)
    end
    
    self.body:setLinearDamping(1)
    self.body:setAngularDamping(1)
    
    if self.originalX then
      self.body:setPosition(self.originalX, self.originalY)
    else
      self.originalX = newCentre.x
      self.originalY = newCentre.y
    end
  elseif not (self.attributes.isBackground or not self.attributes.collidable) then
    if not self.fixture:isDestroyed() then
      self.fixture:destroy()
    end
    self.shape = love.physics.newChainShape(true, self.points)
    self.fixture = love.physics.newFixture(self.body, self.shape)
    self.fixture:setFriction(0.8)
  else
    if not self.fixture:isDestroyed() then
      if self.attributes.isBackground then
        self.type = "backgroundPolygon"
      end
      self.fixture:destroy()
    end
  end
  
  if self.attributes.hasGravity then
    self.body:setGravityScale(1)
  else
    self.body:setGravityScale(0)
  end
end


function LevelPolygon:treatCoordinates(x, y)
  --Snap to an axis or any other changes I wish to make
  
  if love.keyboard.isScancodeDown("lshift") then
    local lastPoint = self.points[self.numberOfPoints]
    --local newPointVector = vector.subtraction(vector.create(x, y), lastPoint)
    
    
    local roundResolution = 50 --math.floor(50/camera.zoom*100)/100
    
    x = math.floor((x-lastPoint.x % roundResolution)/roundResolution+0.5)*roundResolution + lastPoint.x % roundResolution
    y = math.floor((y-lastPoint.y % roundResolution)/roundResolution+0.5)*roundResolution + lastPoint.y % roundResolution
    

    
  end
  return x, y
end


function LevelPolygon:deletePolygon()
  if not self.remove then
    if builder.selectedPolygon == self then
      builder.selectedPolygon = nil
    end
    if not self.attributes.isBackground then
      
      if not self.fixture:isDestroyed() then
        self.fixture:destroy( ) 
      end
    end
    self.body:destroy( )
    self.remove = true
    flushDrawTable()
  end
end


function LevelPolygon:getTopCorner()
  local lowestX = self.points[1]
  local lowestY = self.points[2]

  
  for i, coord in ipairs(self.points) do
    if i > 2 then
      if i%2 == 1 then
        --X
        if coord < lowestX then
          lowestX = coord
        end
      else
        --Y
        if coord < lowestY then
          lowestY = coord
        end
      end
    end
  end
  return vector.create(lowestX, lowestY)
end


function LevelPolygon:averagePoints()
  local average = vector.create(0, 0)
  for i = 1, #self.points, 2 do
    local point = vector.create(self.points[i], self.points[i+1])

    average = vector.addition(point, average)
    
  end
  return vector.scalarDivision(average, #self.points/2)
end


function LevelPolygon:checkSelect()
  return (not self.attributes.ignore) and isPointInPolygon(mouse.worldX, mouse.worldY, {self.body:getWorldPoints(unpack(self.points))})
end


function LevelPolygon:findMiddle(i)
  --Return point exactly between two points
  
  local numberOfPoints = 0
  
  for i, v in ipairs(self.points) do
    numberOfPoints = numberOfPoints +1
  end
  numberOfPoints = numberOfPoints/2
  
  
  local p1 = vector.create(self.points[2*i-1], self.points[2*i])
  local p2 = {}
  if i == numberOfPoints then
    p2 = vector.create(self.points[1], self.points[2])
  else
    p2 = vector.create(self.points[2*(i+1)-1], self.points[2*(i+1)])
  end
  
  
  local returnVector = vector.scalarDivision(vector.addition(p1, p2), 2)
  
  return returnVector.x, returnVector.y
end 



function LevelPolygon:update(dt)
  
  if self.attributes.hasPhysics then
    if self.attributes.springBack then
      local forceVector = vector.create(self.originalX - self.body:getX(), self.originalY - self.body:getY())
      
      self.body:applyForce(forceVector.x*self.body:getMass(), forceVector.y*self.body:getMass())
    end
    if self.attributes.rotateBack then
      
      local torque = -self.body:getAngle()*100000
      
      self.body:applyTorque(torque*self.body:getMass())
    end
    
    if self.attributes.slowMotion then
      
      self:slowMotion(dt, 0.5)
    end
    
  end
end


function LevelPolygon:slowMotion(dt, newSpeed)
  
  local x, y = self.body:getPosition()
  local vX, vY = self.body:getLinearVelocity()
  
  local dx = vX*dt*(1-newSpeed)
  local dy = vY*dt*(1-newSpeed)
  
  print(x - dx, y - dy)
  
  self.body:setPosition(x - dx, y - dy)
  
  
end


function LevelPolygon:updateBuilder()
  
  if builder.mode == "polygon" then
    if self.finalised then
      
      if self.selected then
        if love.keyboard.isScancodeDown("delete") then
          local selectedPoint = self:selectPoint(mouse.worldX, mouse.worldY, self.points)
          if selectedPoint then
            self:removePoint(selectedPoint)
          end
        end
      end
      
      
      local inThisPolygon = isPointInPolygon(mouse.worldX, mouse.worldY, {self.body:getWorldPoints(unpack(self.points))})

      if inThisPolygon then
        --If mouse is over polygon
        if love.keyboard.isScancodeDown("delete") and not self.selected then
          if not self.attributes.protected then
            self:deletePolygon()
          end
        end
      end
      
      if self.selected then
        if love.mouse.isDown(1) then
          
          local selectedPoint = self:selectPoint(mouse.worldX, mouse.worldY)
          if selectedPoint then
            --Selected and finalised
            local transformedX, transformedY = lockToPixels(mouse.worldX, mouse.worldY, "small")
            self:movePoint(transformedX - self.originalX, transformedY - self.originalY, selectedPoint)
          else
            if mouse.left == false then
              --Add a point to a polygon
              local appendPointTable = {}
              for i = 1, #self.points, 2 do
                local x, y = self:findMiddle((i+1)/2)
                table.insert(appendPointTable, x)
                table.insert(appendPointTable, y)
              end
              local selectedAppendPoint = self:selectPoint(mouse.worldX, mouse.worldY, appendPointTable)
              if selectedAppendPoint then
                local i = 2*selectedAppendPoint-1
                local appendX = appendPointTable[i]
                local appendY = appendPointTable[i+1]
                
                table.insert(self.points, i+2, appendX)
                table.insert(self.points, i+3, appendY)
                --Fixes weird move vertex bug
                builder.selectedVertex = selectedAppendPoint+1
                
              end
            end
          end
          mouse.left = true
        else
          mouse.left = false
        end
      end
    end
  end
  
  if self.selected then
    --Selected but not finalised
    builder.selectedPolygon = self
  elseif builder.selectedPolygon == self then
    builder.selectedPolygon = nil
  end
end







function LevelPolygon:draw()
  
  if self.finalised then
    
    if self.attributes.draw then
      local centreX, centreY = nil, nil
      if self.body then
        centreX, centreY = self.body:getWorldCenter()
      end

      self:drawFinalised()
    end
    
    
    
    local worldPoints = {self.body:getWorldPoints(unpack(self.points))}
    
    if self.selected and builderMode then
      love.graphics.setColor(0,0,1)
      blue.graphics.polygon("line", {self.body:getWorldPoints(unpack(self.points))})
    end
    if debugMode or (self.selected and builderMode) then
      for i = 1, #self.points, 2 do
        if i == 1 then
          love.graphics.setColor(0,0,0)
        else
          love.graphics.setColor(1,0,0)
        end
        
        self.originalX = self.originalX or 0
        self.originalY = self.originalY or 0
        
        blue.graphics.circle("line", self.points[i] + self.originalX, self.points[i+1] + self.originalY, 2/camera.zoom)
        if self.selected and builderMode then
          love.graphics.setColor(1,0,0, 0.3)
          blue.graphics.circle("line", self.points[i] + self.originalX, self.points[i+1] + self.originalY, 15/camera.zoom)
        end
      end
      
      
      
      
      
      if self.triangleShapes and self.triangleShapes[1] then
        for i, triangle in ipairs(self.triangleShapes) do
          
          local shapePoints = {self.body:getWorldPoints(triangle:getPoints())}
          for i = 1, #shapePoints, 2 do
            if i == 1 then
              love.graphics.setColor(0,0,0)
            else
              love.graphics.setColor(0,1,1)
            end
            blue.graphics.circle("line", shapePoints[i], shapePoints[i+1], 5/camera.zoom)
            
          end
        end
      end
      
      
      
      if self.selected and builderMode then
        for i = 1, #self.points, 2 do
          local x, y = self:findMiddle((i+1)/2)
          love.graphics.setColor(0,1,0)
          x = x + self.originalX
          y = y + self.originalY
          blue.graphics.circle("line", x, y, 2/camera.zoom)
        end
      end
    end
    
    if debugMode then
      if not self.attributes.draw then
        love.graphics.setColor(1,0,0)
        blue.graphics.polygon("line", {self.body:getWorldPoints(unpack(self.points))})
      end
    
      love.graphics.setColor(0,1,1)
      blue.graphics.circle("fill", self.body:getX(), self.body:getY(), 4/camera.zoom)

    end
    

    
  else
    local lastPoint = false
    local firstPoint = false
    
    for i, point in ipairs(self.points) do
      love.graphics.setColor(0,0,0)
      blue.graphics.circle("line", point.x, point.y, 3/camera.zoom)
      if i ~= 1 then
        love.graphics.setColor(0.8,0.2,0.2)
        blue.graphics.line(point.x, point.y, lastPoint.x, lastPoint.y)
      else
        firstPoint = point
      end
      lastPoint = point
    end
    if self.numberOfPoints > 2 then
      love.graphics.setColor(0.2,0.2,1,0.2)
      blue.graphics.line(lastPoint.x, lastPoint.y, firstPoint.x, firstPoint.y)
    end
    
    lastPoint = self.points[self.numberOfPoints]
    local mouseX, mouseY = self:treatCoordinates(mouse.worldX, mouse.worldY)
    
    love.graphics.setColor(0,0,1,0.5)
    blue.graphics.line(mouseX, mouseY, lastPoint.x, lastPoint.y)
  end
end

